"""
Parser for nominet whois data formats
"""

import re
from datetime import datetime, date
from seine.whois.formats import WhoisDataParser, WhoisError

SECTION_HEADERS = {
    'Domain name':          'domainname',
    'Registrant':           'registrant',
    'Trading as':           'trading_as',
    'Registrant type':      'registrant_type',
    "Registrant's address": 'registrant_address',
    'Registrar':            'registrar',
    'Relevant dates':       'dates',
    'Registration status':  'status',
    'Name servers':         'nameservers',
}

DATE_FIELD_MAP = {
    'Registered on:':   'created',
    'Renewal date:':    'expires',
    'Last updated:':    'modified',
}
DATE_PARSER = lambda value: datetime.strptime(value, '%d-%b-%Y').date()
# Registrations before good bookkeeping
OLD_DATE_BANNER = 'before Aug-1996'
OLD_DATE_VALUE = date(1996,8,01)

RE_NS_LIST = re.compile('(?P<ns>.*) (?P<addresses>.*)$')
END_HEADER_PREFIX = 'WHOIS lookup made at'

class nominet(WhoisDataParser):
    tlds = ( 'uk', )

    def parse(self, domain, data):
        """Parse data

        Parse Nominet whois data

        """

        def next_section(name, section, value):
            if section is not None and value is not None:
                self.set(section, value)
            return (name, None)

        def push_value(value, new_value):
            if value is None:
                value = new_value
            else:
                if type(value) != list:
                    value = [value]
                value.append(new_value)
            return value

        data = WhoisDataParser.parse(self, domain, data)

        section = None
        value = None
        for l in [l.strip() for l in data]:
            if l.startswith('% ') or l == '':
                continue
            l = l.decode('utf-8')

            if l[:len(END_HEADER_PREFIX)] == END_HEADER_PREFIX:
                break

            if l[-1]==':' and l[:-1] in SECTION_HEADERS.keys():
                (section, value) = next_section(SECTION_HEADERS[l[:-1]], section, value)
                continue

            if section == 'dates':
                for key, field in DATE_FIELD_MAP.items():
                    if l[:len(key)] == key:
                        datevalue = l[len(key):].strip()
                        if datevalue != OLD_DATE_BANNER:
                            datevalue =  DATE_PARSER(datevalue)
                        else:
                            datevalue = OLD_DATE_VALUE
                        self.set(field, datevalue)
                        break

            elif section == 'nameservers':
                m = RE_NS_LIST.match(l)
                if m:
                    ns = m.groupdict()['ns'].strip()
                    glue = [x.strip() for x in m.groupdict()['addresses'].split(',')]
                    self.set('glue_%s' % ns, glue)
                else:
                    ns = l.strip()
                value = push_value(value, ns)

            else:
                value = push_value(value, l)

        if section is not None and value is not None:
            self.set(section, value)




#!/usr/bin/env python
"""
Generic SNMP device process listing monitoring classes
"""

import os,re,logging
from configobj import ConfigObj

import pysnmp.error as PySNMPError
from pysnmp.proto import rfc1902

from seine.snmp import SNMPError
from seine.snmp.client import SNMPClient
from seine.address import IPv4Address,EthernetMACAddress

IF_STATUS_UP = 1 
IF_STATUS_DOWN = 2

IF_STATUS_CODES = [ None, IF_STATUS_UP, IF_STATUS_DOWN ]

PROCESS_OIDS_MAP = {
    'index': { 
        'oid': '.1.3.6.1.2.1.25.4.2.1.1',
        'decode': lambda oid,x: (oid,int(x)), 
    },
    'name': {
        'oid': '.1.3.6.1.2.1.25.4.2.1.2',
        'decode': lambda oid,x: (oid,str(x)),
    },
    'args': {
        'oid': '1.3.6.1.2.1.25.4.2.1.5',
        'decode': lambda oid,x: (oid,str(x)),
    },
    'mem': {
        'oid': '.1.3.6.1.2.1.25.5.1.1.2', 
        'decode': lambda oid,x: (oid,int(x)),
    },
    'cputime': { 
        'oid': '.1.3.6.1.2.1.25.5.1.1.1', 
        'decode': lambda oid,x: (oid,int(x)),
    },
    'runstate': { 
        'oid': '.1.3.6.1.2.1.25.4.2.1.7', 
        'decode': lambda oid,x: (oid,int(x)), 
    },
}

class SNMPProcessIndexCache(dict):
    def __init__(self,path=None):
        self.path = path
        self.load()

    def load(self):
        if self.path is None or not os.path.isfile(self.path):
            return
        for host,opts in ConfigObj(self.path).items():
            self[host] = dict(opts)

    def save(self):
        if self.path is None:   
            return
        c = ConfigObj()
        for host,opts in self.items():
            if len(opts)==0:
                continue
            if not c.has_key(host):
                c[host] = {}
            for index,details in opts.items():
                index = str(index)
                c[host][index] = details
        c.write(open(self.path,'w'))

class SNMPProcessStatus(SNMPClient):
    """
    Cacheable dictionary of process details: pid, name, args mem, cputime 
    """
    def __init__(self,oid_map=PROCESS_OIDS_MAP,*args,**kwargs):
        SNMPClient.__init__(self,oid_map=PROCESS_OIDS_MAP,*args,**kwargs)

    def __getattr__(self,attr):
        if attr in PROCESS_OIDS_MAP.keys():
            d = PROCESS_OIDS_MAP[attr]
            oid = d['oid']
            return [d['decode'](k,v) for k,v in self.walk(oid).items()]
        raise AttributeError('No such SNMPProcessStatus attribute: %s' % attr)

    def update_indexes(self):
        self.indexes = {}
        self.log.info('Loading indexes')
        for section in ['index','name','args','mem','cputime']:
            for k,v in getattr(self,section):
                index = int(k.split('.')[-1])
                if not self.indexes.has_key(index):
                    self.indexes[index] = {}
                if section in ['mem','cputime','index']:
                    v = int(v)
                self.indexes[index][section] = v
        self.index_cache[self.address] = dict(self.indexes)
        self.index_cache.save()




#!/usr/bin/env python
"""
Parser for saudinic whois data formats
"""

import re
from datetime import datetime
from seine.whois.formats import WhoisDataParser, WhoisError

FIELD_MAP = {
    'Domain Name':      'domainname',
    'descr':            'owner',
    'address':          'address',
    'phone':            'telephone',
    'status':           'status',
    'nserver':          'nameservers',
    'created':          'created',
    'modified':         'updated',
    'expires':          'expires',
    'dnssec':           'dnssec',
}

RE_NS_LIST = re.compile('(?P<ns>.*) \((?P<addresses>.*)\)$')

DATE_FIELDS = ( 'updated', 'created', 'expires', )
DATE_PARSER = lambda value: datetime.strptime(value, '%d.%m.%Y').date()


class saudinic(WhoisDataParser):
    tlds = ( 'sa', )

    def parse(self, domain, data):
        """Parse data

        Parse SaudiNIC whois data

        """
        def next_section(name, section, value):
            if section is not None and value is not None:
                self.set(section, value)
            return (name,None)

        def push_value(value, new_value):
            if value is None:
                value = new_value
            else:
                if type(value) != list:
                    value = [value]
                value.append(new_value)
            return value

        data = WhoisDataParser.parse(self, domain, data)

        section = None
        value = None
        for l in [l.strip() for l in data]:
            if l.startswith('% ') or l == '':
                continue
            l = l.decode('utf-8')

            if l[:12] == 'Domain Name:':
                self.set('domainname', l[12:].lstrip())
                continue

            if l == 'Registrant:':
                (section, value) = next_section('registrant', section, value)
                continue

            if l == 'Administrative Contact:':
                (section, value) = next_section('admin_contact', section, value)
                continue

            if l == 'Technical Contact:':
                (section, value) = next_section('technical_contact', section, value)
                continue

            if l == 'Name Servers:':
                (section, value) = next_section('nameservers', section, value)
                continue

            if l[:11] == 'Created on:':
                self.set('created', datetime.strptime(l[12:], '%Y-%m-%d').date())
                continue
            if l[:16] == 'Last Updated on:':
                self.set('modified', datetime.strptime(l[17:], '%Y-%m-%d').date())
                continue

            if section == 'registrant':
                if l[:8] == 'Address:':
                    (section, value) = next_section('registrant_address', section, value)
                    value = l[9:]
                    continue
                else:
                    value = push_value(value, l)

            if section == 'admin_contact':
                if l[:8] == 'Address:':
                    (section, value) = next_section('admin_address', section, value)
                    value = l[9:]
                    continue
                else:
                    value = push_value(value, l)

            if section == 'technical_contact':
                if l[:8] == 'Address:':
                    (section, value) = next_section('technical_address', section, value)
                    value = l[9:]
                    continue
                else:
                    value = push_value(value, l)

            if section in ( 'registrant_address', 'admin_address', 'technical_address', ):
                value = push_value(value, l)

            elif section == 'nameservers':
                m = RE_NS_LIST.match(l)
                if m:
                    ns = m.groupdict()['ns'].strip()
                    glue = [x.strip() for x in m.groupdict()['addresses'].split(',')]
                    self.set('glue_%s' % ns, glue)
                else:
                    ns = l.strip()

                value = push_value(value, ns)

        if section is not None and value is not None:
            self.set(section, value)




"""
Module to query the DNS zone NS server delegation records starting
from DNS root servers.
"""

import random

from seine.dns import DNSError
from seine.address import IPv4Address, IPv6Address
from seine.dns.resolver import resolve_records
from seine.dns.rootservers import RootServers

class SOA(object):
    def __init__(self, server, domain):
        self.server = server
        self.domain = domain

        res = resolve_records(self.domain, self.server, 'SOA')
        if len(res['results']) != 1:
            raise ValueError('Invalid SOA query result: %s' % res['results'])
        data = res['results'][0]

        if data['rrtype'] != 'SOA':
            raise ValueError('Response is not a SOA record')

        name = '%s' % data['name']
        if self.domain != name:
            raise ValueError('Received SOA name %s expected %s' % (name, self.domain))

        self.master_server = '%s' % data['mname']
        self.rname = '%s' % data['rname']

        for field in ('serial', 'refresh', 'retry', 'expire', 'minimum'):
            if field not in data.keys():
                raise ValueError('Invalid SOA query result: missing %s' % field)
            setattr(self, field, data[field])

    def __repr__(self):
        return '%s serial %d refresh %d retry %d expire %d minimum %d contact %s' % (
            self.domain,
            self.serial,
            self.refresh,
            self.retry,
            self.expire,
            self.minimum,
            self.rname
        )

    def __cmp__(self, other):
        """Compare SOA records

        Please note we do NOT compare server from which the record originates from: identical
        SOA records are expected from multiple sources

        """
        if not isinstance(other, SOA):
            raise ValueError('Compared target must be SOA instance: %s' % type(other))

        if self.master_server != other.master_server:
            return cmp(self.master_server, other.master_server)

        if self.domain != other.domain:
            return cmp(self.domain, other.domain)

        if self.rname != other.rname:
            return cmp(self.rname, other.rname)

        for field in ('serial', 'refresh', 'retry', 'expire', 'minimum'):
            a = getattr(self, field)
            b = getattr(other, field)
            if a != b:
                return cmp(a, b)

        return 0

    def __eq__(self, other):
        return self.__cmp__(other) == 0

    def __ne__(self, other):
        return self.__cmp__(other) != 0

    def __lt__(self, other):
        return self.__cmp__(other) < 0

    def __lte__(self, other):
        return self.__cmp__(other) <= 0

    def __gt__(self, other):
        return self.__cmp__(other) > 0

    def __gte__(self, other):
        return self.__cmp__(other) >= 0

    @property
    def email(self):
        try:
            address, domain = self.rname.split('.', 1)
            return '%s@%s' % (address, domain.rstrip('.'))
        except ValueError:
            return None

    def validate(self):
        """Validate SOA

        Validate time values in SOA record

        """
        errors = []
        warnings = []
        if self.refresh < 0:
            errors.append('Invalid refresh value: %s' % refresh)

        if self.email is None:
            warnings.append('Error parsing email address from %s' % self.rname)

        if self.expire <= self.refresh:
            errors.append('Zone expire <= refresh: %s %s' % (self.expire, self.refresh))

        if self.expire <= self.retry:
            errors.append('Zone expire <= retry: %s %s' % (self.expire, self.retry))

        if self.refresh < self.minimum:
            warnings.append('Zone refresh time is smaller than minimum TTL: %s %s' % (self.refresh, self.minimum))

        return errors, warnings

class DelegateServer(object):
    def __init__(self, delegation, rrtype, name, address):
        self.delegation = delegation
        self.rrtype = rrtype
        self.name = name
        self.address = address

    def __repr__(self):
        return self.hostname

    @property
    def hostname(self):
        return '%s' % self.name.__str__().rstrip('.')

    @property
    def domain(self):
        return '%s.' % self.delegation.domain

    @property
    def soa(self):
        try:
            return SOA(self.address, self.domain)
        except ValueError, emsg:
            raise DNSError('Error retrieving SOA for %s from %s: %S' % (self.address, self.domain, emsg))

    @property
    def nameservers(self):
        addresses = []
        res = resolve_records(self.domain, self.address, 'NS')
        for r in res['results']:
            if r['rrtype'] != 'NS':
                continue

            server = '%s' % r['target']
            addresses.append(server.rstrip('.'))

        return addresses

class DNSZoneDelegation(list):
    """
    Class to fetch the DNS zone NS delegation details from root servers
    and from the TLD servers for given domain.

    Note that 'rootfile' must refere to a valid copy of db.root root
    server list. You can fetch a local copy from official sources with
    update() method of netutils.dnstools.rootservers.RootServers().
    """
    def __init__(self, domain, rrtype='A', rootfile='/tmp/db.root', timeout=5):
        """
        This class only checks for NS delegations with given rrtype: i.e.,
        if you want to check both IPv4 and IPv6 delegation, call this twice
        with rrtype='A' and rrtype='AAAA'. Don't change the attribute on the
        fly, it may break something.
        """
        self.domain = domain
        self.tld = '%s.' % self.domain.split('.')[-1]

        self.rootfile = rootfile
        self.timeout = int(timeout)

        if rrtype == 'A':
            self.address_type = 'IPv4'
        elif rrtype == 'AAAA':
            self.address_type = 'IPv6'
        else:
            raise ValueError('Invalid RR type: %s' % rrtype)

        self.rrtype = rrtype
        self.tld_servers = []

    def __str__(self):
        return '%s servers: %s' % (
            self.address_type,
            ','.join(self.delegated_servers),
        )

    @property
    def delegated_servers(self):
        return [s.hostname for s in self]

    def update_tld_rootservers(self):
        """
        Find the DNS TLD servers for the domain from root servers
        """

        self.tld_servers = []
        rootservers = RootServers(self.rootfile)
        if not rootservers.is_downloaded:
            rootservers.download()

        server = rootservers.random_rootserver(self.rrtype).address
        rs = resolve_records(query=self.tld, nameserver=server, rrtype='NS', timeout=self.timeout)

        tld_server_addresses = {}
        for entry in filter(lambda x: x['rrtype']==self.rrtype, rs['additional']):
            try:
                tld_server_addresses[entry['name']] = entry['address']
            except IndexError:
                raise DNSError('Additional data did not contain proper %s records' % self.rrtype)

        self.tld_servers = []
        for entry in filter(lambda x: x['rrtype']=='NS', rs['authority']):
            try:
                self.tld_servers.append(tld_server_addresses[entry['target']])
            except IndexError:
                continue

    def update_delegations(self):
        """
        Query zone delegation details from one of the TLD servers
        """
        self.__delslice__(0, len(self))

        self.update_tld_rootservers()
        if len(self.tld_servers) == 0:
            raise DNSError('No TLD servers for %s' % self.domain)
        server = self.tld_servers[random.randint(0, len(self.tld_servers)-1)]

        rs = resolve_records( self.domain, server, 'NS', self.timeout)

        reg_server_addresses = {}
        for entry in filter(lambda x: x['rrtype']==self.rrtype, rs['additional']):
            reg_server_addresses[entry['name']] = entry['address']

        for entry in filter(lambda x: x['rrtype']=='NS', rs['authority']):
            try:
                address = reg_server_addresses[entry['target']]
            except IndexError:
                # Need to query address for server
                address = None

            self.append(DelegateServer(self, self.rrtype, entry['target'], address ))



#!/usr/bin/env python
"""
Generic SNMP device monitoring classes
"""

import os,re,logging
from configobj import ConfigObj

import pysnmp.error as PySNMPError
from pysnmp.proto import rfc1902

from seine.snmp import SNMPError
from seine.snmp.client import SNMPClient
from seine.address import IPv4Address,EthernetMACAddress

IF_STATUS_UP = 1 
IF_STATUS_DOWN = 2
IF_STATUS_CODES = [ None, IF_STATUS_UP, IF_STATUS_DOWN ]

INTERFACE_IPV4_OID_PREFIX = '.1.3.6.1.2.1.3.1.1.3'

NETWORK_OIDS_MAP = {
    'index': { 
        'oid': '.1.3.6.1.2.1.2.2.1.1', 
        'decode': lambda oid,x: (oid,int(x)), 
    },
    'mtu': {
        'oid': '.1.3.6.1.2.1.2.2.1.4', 
        'decode': lambda oid,x: (oid,int(x)),
    },
    'speed': {
        'oid': '.1.3.6.1.2.1.2.2.1.5', 
        'decode': lambda oid,x: (oid,int(x)),
    },
    'phys_address': {
        'oid': '.1.3.6.1.2.1.2.2.1.6', 
        'decode': lambda oid,x: (oid,EthernetMACAddress(x)),
    },
    'description': { 
        'oid': '.1.3.6.1.2.1.2.2.1.2', 
        'decode': lambda oid,x: (oid,str(x)),
    },
    'oper_status': { 
        'oid': '.1.3.6.1.2.1.2.2.1.8', 
        'decode': lambda oid,x: (oid,IF_STATUS_CODES[int(x)]), 
    },
    'admin_status': { 
        'oid': '.1.3.6.1.2.1.2.2.1.7', 
        'decode': lambda oid,x: (oid,IF_STATUS_CODES[int(x)]), 
    },
    'in_octets': { 
        'oid': '.1.3.6.1.2.1.2.2.1.10', 
        'decode': lambda oid,x: (oid,long(x)),
    },
    'in_ucast_pkts': { 
        'oid': '.1.3.6.1.2.1.2.2.1.11', 
        'decode': lambda oid,x: (oid,long(x)),
    },
    'in_non_ucast_pkts': { 
        'oid': '.1.3.6.1.2.1.2.2.1.12', 
        'decode': lambda oid,x: (oid,long(x)),
    },
    'in_discards': { 
        'oid': '.1.3.6.1.2.1.2.2.1.13', 
        'decode': lambda oid,x: (oid,long(x)),
    },
    'in_errors': { 
        'oid': '.1.3.6.1.2.1.2.2.1.14', 
        'decode': lambda oid,x: (oid,long(x)),
    },
    'in_unknown_proto': { 
        'oid': '.1.3.6.1.2.1.2.2.1.15', 
        'decode': lambda oid,x: (oid,long(x)),
    },
    'out_octets': { 
        'oid': '.1.3.6.1.2.1.2.2.1.16', 
        'decode': lambda oid,x: (oid,long(x)),
    },
    'out_ucast_pkts': { 
        'oid': '.1.3.6.1.2.1.2.2.1.17', 
        'decode': lambda oid,x: (oid,long(x)),
    },
    'out_non_ucast_pkts': { 
        'oid': '.1.3.6.1.2.1.2.2.1.18', 
        'decode': lambda oid,x: (oid,long(x)),
    },
    'out_discards': { 
        'oid': '.1.3.6.1.2.1.2.2.1.19', 
        'decode': lambda oid,x: (oid,long(x)),
    },
    'out_errors': { 
        'oid': '.1.3.6.1.2.1.2.2.1.20', 
        'decode': lambda oid,x: (oid,long(x)),
    },
    'out_qlen': { 
        'oid': '.1.3.6.1.2.1.2.2.1.21', 
        'decode': lambda oid,x: (oid,long(x)),
    },
    'name': { 
        'oid': '.1.3.6.1.2.1.31.1.1.1.1',
        'decode': lambda oid,x: (oid,str(x)),
    },
    'in_mcast_pkts': { 
        'oid': '.1.3.6.1.2.1.31.1.1.1.2',
        'decode': lambda oid,x: (oid,long(x)),
    },
    'in_bcast_pkts': { 
        'oid': '.1.3.6.1.2.1.31.1.1.1.3',
        'decode': lambda oid,x: (oid,long(x)),
    },
    'out_mcast_pkts': { 
        'oid': '.1.3.6.1.2.1.31.1.1.1.4',
        'decode': lambda oid,x: (oid,long(x)),
    },
    'out_bcast_pkts': { 
        'oid': '.1.3.6.1.2.1.31.1.1.1.5',
        'decode': lambda oid,x: (oid,long(x)),
    },
    'in_64bit_octets': { 
        'oid': '.1.3.6.1.2.1.31.1.1.1.6',
        'decode': lambda oid,x: (oid,long(x)),
    },
    'in_64bit_ucast_pkts': { 
        'oid': '.1.3.6.1.2.1.31.1.1.1.7',
        'decode': lambda oid,x: (oid,long(x)),
    },
    'in_64bit_mcast_pkts': { 
        'oid': '.1.3.6.1.2.1.31.1.1.1.8',
        'decode': lambda oid,x: (oid,long(x)),
    },
    'in_64bit_bcast_pkts': { 
        'oid': '.1.3.6.1.2.1.31.1.1.1.9',
        'decode': lambda oid,x: (oid,long(x)),
    },
    'out_64bit_octets': { 
        'oid': '.1.3.6.1.2.1.31.1.1.1.10',
        'decode': lambda oid,x: (oid,long(x)),
    },
    'out_64bit_ucast_pkts': { 
        'oid': '.1.3.6.1.2.1.31.1.1.1.11',
        'decode': lambda oid,x: (oid,long(x)),
    },
    'out_64bit_mcast_pkts': { 
        'oid': '.1.3.6.1.2.1.31.1.1.1.12',
        'decode': lambda oid,x: (oid,long(x)),
    },
    'out_64bit_bcast_pkts': { 
        'oid': '.1.3.6.1.2.1.31.1.1.1.13',
        'decode': lambda oid,x: (oid,long(x)),
    },
    'updown_trap_enable': { 
        'oid': '.1.3.6.1.2.1.31.1.1.1.14',
        'decode': lambda oid,x: (oid,int(x)== 1 and True or False),
    },
    'speed_fastlink': { 
        'oid': '.1.3.6.1.2.1.31.1.1.1.15',
        'decode': lambda oid,x: (oid,long(x)),
    },
    'promiscuous': { 
        'oid': '.1.3.6.1.2.1.31.1.1.1.16',
        'decode': lambda oid,x: (oid,int(x)== 1 and True or False),
    },
    'connector_present': { 
        'oid': '.1.3.6.1.2.1.31.1.1.1.17',
        'decode': lambda oid,x: (oid,int(x)== 1 and True or False),
    },
    'alias': { 
        'oid': '.1.3.6.1.2.1.31.1.1.1.18',
        'decode': lambda oid,x: (oid,str(x)),
    },
}

class SNMPNetworkInterfaces(SNMPClient):
    def __init__(self,*args,**kwargs):
        SNMPClient.__init__(self,oid_map=NETWORK_OIDS_MAP,*args,**kwargs)

    def update_indexes(self):
        self.indexes = {}
        self.log.info('Loading indexes')
        for section in ['index','name','description']:
            for k,v in getattr(self,section):
                index = int(k.split('.')[-1])
                if not self.indexes.has_key(index):
                    self.indexes[index] = {}
                self.indexes[index][section] = v
        self.index_cache[self.address] = self.indexes
        self.index_cache.save()

    def set_alias(self,ifindex,value):
        try:
            ifindex = int(ifindex)
            if ifindex not in self.indexes.keys():
                raise ValueError
        except ValueError:
            raise ValueError('Invalid ifindex')
        oid = '.'.join([NETWORK_OIDS_MAP['alias']['oid'],str(ifindex)])
        print oid
        try:
            value = rfc1902.OctetString(value)
            self.set(oid,value)
        except SNMPError,emsg:
            raise ValueError('Error updating alias: %s' % emsg)

    def interface_ipv4_addresses(self,ifindex):
        if_oid = '.'.join([INTERFACE_IPV4_OID_PREFIX,str(ifindex)])
        addresses = []
        for (oid,value) in self.walk(if_oid).items():
            try:
                addresses.append(IPv4Address(value))
            except ValueError:
                raise ValueError('Invalid IPv4 address from OID %s' % oid)
        return addresses  

    def interface_names(self):
        if self.indexes == {}:
            self.update_indexes()
        return self.indexes

    def interface_details(self,ifindex,fields=None):
        if self.indexes == {}:
            self.update_indexes()
        try:
            ifindex = int(ifindex)
            if ifindex not in self.indexes.keys():
                raise ValueError
        except ValueError:
            raise ValueError('Invalid ifindex value %s' % ifindex)
        details = {}
        if fields is not None:
            for f in fields:
                if f not in NETWORK_OIDS_MAP.keys():
                    raise ValueError('Invalid field name %s' % f)
        else:
            fields = NETWORK_OIDS_MAP.keys()
        for k in fields:
            d = NETWORK_OIDS_MAP[k]
            oid = '.'.join([d['oid'],str(ifindex)])
            try:
                (oid,value) = self.get(oid)
                try:
                    details[k] = d['decode'](oid,value)[1]
                except AttributeError:
                    details[k] = None
            except ValueError:
                raise ValueError('Error parsing oid %s' % oid)
        if fields is None:
            details['ipv4_addresses'] = self.interface_ipv4_addresses(ifindex)
        return details

    def interface_status(self,ifindexes=None):
        if self.indexes == {}:
            self.update_indexes()
        fields = ['name','description','alias','oper_status','admin_status']
        if ifindexes is None:
            details = self.indexes()
            for k in fields:
                for oid,value in getattr(self,k):
                    index = int(oid.split('.')[-1])
                    details[index][k] = value
        else:
            if type(ifindexes) != list: ifindexes = [ifindexes]
            details = {}
            for ifindex in ifindexes:
                try:
                    details[ifindex] = self.interface_details(ifindex,fields)
                except ValueError,emsg:
                    raise ValueError(emsg)
        return details

    def interface_octet_counters(self,ifindexes=None):
        if self.indexes == {}:
            self.update_indexes()
        fields =  ['in_octets','out_octets','in_64bit_octets','out_64bit_octets']
        if ifindexes is None:
            details = self.indexes()
            for k in fields:
                for oid,value in getattr(self,k):
                    index = int(oid.split('.')[-1])
                    details[index][k] = value
        else:
            if type(ifindexes) != list: ifindexes = [ifindexes]
            details = {}
            for ifindex in ifindexes:
                details[ifindex] = self.interface_details(ifindex,fields)
        return details

    def interface_packet_counters(self,ifindexes=None):
        if self.indexes == {}:
            self.update_indexes()
        fields = [
            'in_ucast_pkts','in_non_ucast_pkts',
            'out_ucast_pkts','out_non_ucast_pkts' 
        ]
        if ifindexes is None:
            details = self.indexes()
            for k in fields:
                for oid,value in getattr(self,k):
                    index = int(oid.split('.')[-1])
                    details[index][k] = value
        else:
            if type(ifindexes) != list: ifindexes = [ifindexes]
            details = {}
            for ifindex in ifindexes:
                details[ifindex] = self.interface_details(ifindex,fields)
        return details




"""
Parse network interfaces and routing tables
"""

import os
import sys
import fnmatch

from systematic.log import Logger, LoggerError

from seine.address import IPv4Address, IPv6Address, \
                          EthernetMACAddress, FirewireLocalLinkAddress


class Interfaces(object):
    """
    Thin wrapper to load OS specific implementation for network interfaces
    """
    __loader = None

    def __init__(self):
        if Interfaces.__loader is None:
            if sys.platform[:5] == 'linux':
                from seine.platform.linux.network import LinuxNetworkInterfaces
                Interfaces.__loader = LinuxNetworkInterfaces()

            elif sys.platform == 'darwin' or fnmatch.fnmatch(sys.platform, 'freebsd*'):
                from seine.platform.bsd.network import BSDNetworkInterfaces
                Interfaces.__loader = BSDNetworkInterfaces()

            else:
                raise ValueError('Interfaces loader for OS not available: %s' % sys.platform)

        self.__dict__['_Interfaces__loader'] = Interfaces.__loader
        self.__loader.update()
        self.__loader.sort()

    def __repr__(self):
        return '%s' % [x.name for x in self.__loader]

    def __getattr__(self, attr):
        return getattr(self.__loader, attr)

    def __getitem__(self, item):
        return self.__loader.__getitem__(item)

    def __iter__(self):
        return self.__loader.__iter__()

    def next(self):
        return self.__loader.next()


class ARP(object):
    __loader = None

    def __init__(self):
        if ARP.__loader is None:
            if sys.platform[:5] == 'linux':
                from seine.platform.linux.network import LinuxARP
                ARP.__loader = LinuxARP()

            elif sys.platform == 'darwin' or fnmatch.fnmatch(sys.platform, 'freebsd*'):
                from seine.platform.bsd.network import BSDARP
                ARP.__loader = BSDARP()

            else:
                raise ValueError('ARP loader for OS not available: %s' % sys.platform)

        self.__dict__['_ARP__loader'] = ARP.__loader
        self.__loader.update()
        self.__loader.sort()

    def __getattr__(self, attr):
        return getattr(self.__loader, attr)

    def __getitem__(self, item):
        return self.__loader.__getitem__(item)

    def __iter__(self):
        return self.__loader.__iter__()

    def next(self):
        return self.__loader.next()

    def match(self, value):
        matches = []
        for entry in self:
            if entry.ethernet is not None and fnmatch.fnmatch('%s' % entry.ethernet, value):
                matches.append(entry)

            if fnmatch.fnmatch('%s' % entry.address, value):
                matches.append(entry)

        return matches


class Routes(object):
    """
    Thin wrapper to load OS specific implementation for network interfaces
    """
    __loader = None

    def __init__(self):
        if Routes.__loader is None:
            if sys.platform[:5] == 'linux':
                from seine.platform.linux.network import LinuxRoutes
                Routes.__loader = LinuxRoutes()

            elif sys.platform == 'darwin' or fnmatch.fnmatch(sys.platform, 'freebsd*'):
                from seine.platform.bsd.network import BSDRoutes
                Routes.__loader = BSDRoutes()

            else:
                raise ValueError('Routes loader for OS not available: %s' % sys.platform)

        self.__dict__['_Routes__loader'] = Routes.__loader
        self.__loader.update()

    def __getattr__(self, attr):
        return getattr(self.__loader, attr)



#!/usr/bin/env python
"""
Common wrapper for snmp command line query scripts
"""

import sys,os,logging

from systematic.shell import Script,ScriptError
from seine.snmp import SNMPError,SNMP_VERSIONS
from seine.snmp.client import SNMPClient,SNMPv1Auth,SNMPv2cAuth,SNMPv3Auth
from seine.snmp.devices.process import SNMPProcessStatus

class SNMPScript(Script):
    def __init__(self,oid_map={},index_cache_path=None):
        Script.__init__(self)

        self.oid_map = oid_map
        self.add_argument('-H','--host',help='SNMP host to connect')
        self.add_argument('-p','--port',type=int,default=161,help='SNMP server port')
        self.add_argument('-1',dest='snmpv1',action='store_true',help='Use SNMP v1')
        self.add_argument('-2',dest='snmpv2c',action='store_true',help='Use SNMP v2c')
        self.add_argument('-3',dest='snmpv3',action='store_true',help='Use SNMP v3')
        self.add_argument('-C','--community',help='SNMP v1/v2c community')
        self.add_argument('-U','--username',help='SNMP v3 username')
        self.add_argument('-a','--authProtocol',help='SNMP v3 auth protocol')
        self.add_argument('-A','--authPass',help='SNMP v3 auth password')
        self.add_argument('-x','--privProtocol',help='SNMP v3 priv protocol')
        self.add_argument('-X','--privPass',help='SNMP v3 priv password')
        self.add_argument('-t','--timeout',type=int,default=1,help='SNMP query timeout')
        self.add_argument('-r','--retries',type=int,default=5,help='SNMP query retries')
        self.add_argument('-f','--cache-path',default=index_cache_path,help='Result index cache path')

    @property
    def client_kwargs(self):
        return {
            'address': self.address, 'port': self.port, 'auth': self.auth,
            'timeout': self.timeout, 'retries': self.retries,
            'index_cache_path': self.index_cache_path,
        }

    def parse_args(self):
        args = Script.parse_args(self)
        self.address = args.host
        self.port = args.port
        self.timeout = args.timeout
        self.retries = args.retries
        self.index_cache_path = args.cache_path

        if args.snmpv3 and args.username and args.authPass:
            self.auth = SNMPv3Auth(args.username,args.authPass,args.privPass,args.authProtocol,args.privProtocol)
        elif args.snmpv2c and args.community:
            self.auth = SNMPv2cAuth(args.community)
        elif args.snmpv1 and args.community:
            self.auth = SNMPv1Auth(args.community)
        else:
            raise ScriptError('No SNMP authentication method provided')

        return args






"""
SNMP client classes to simplify using pysnmp
"""

import sys
import os
import socket
from configobj import ConfigObj

from systematic.log import Logger, LoggerError

from pyasn1.error import PyAsn1Error, SubstrateUnderrunError, ValueConstraintError
from pysnmp.carrier.error import CarrierError

from pysnmp.entity import config as snmpconfig
from pysnmp.entity.rfc3413.oneliner.cmdgen import CommunityData, UsmUserData
from pysnmp.entity.rfc3413.oneliner.cmdgen import UdpTransportTarget
from pysnmp.entity.rfc3413.oneliner.cmdgen import CommandGenerator

from seine.snmp import cmp_oid, SNMPError, SNMP_VERSIONS

DEFAULT_AGENT = 'seine'
DEFAULT_PORT = 161
DEFAULT_TIMEOUT = 1
DEFAULT_RETRIES = 5
DEFAULT_VERSION='2c'
DEFAULT_V3_AUTH_PROTOCOL = 'SHA'
DEFAULT_V3_PRIV_PROTOCOL = 'AES'

V3_AUTH_PROTOCOLS = {
    'MD5': snmpconfig.usmHMACMD5AuthProtocol,
    'SHA': snmpconfig.usmHMACSHAAuthProtocol,
    'NOAUTH': snmpconfig.usmNoAuthProtocol,
}
V3_PRIV_PROTOCOLS = {
    'DES': snmpconfig.usmDESPrivProtocol,
    'AES': snmpconfig.usmAesCfb128Protocol,
    'NOPRIV': snmpconfig.usmNoPrivProtocol,
}

class SNMPAuth(object):
    """SNMP authentication base class

    Base class for SNMP*Auth classes

    """
    def __init__(self, version):
        if version not in SNMP_VERSIONS:
            raise SNMPError('Invalid SNMP version: %s' % version)
        self.version = version
        self.auth = None

class SNMPv1Auth(SNMPAuth):
    """SNMP v1 community authentication

    Authenticate with SNMP v1 community

    """
    def __init__(self, community, securityname=DEFAULT_AGENT):
        SNMPAuth.__init__(self, '1')
        self.community = community
        self.securityname = securityname
        self.auth = CommunityData(self.securityname, self.community, 0)

    def __repr__(self):
        return '%s SNMP v1 community %s' % (self.securityname, self.community)

class SNMPv2cAuth(SNMPAuth):
    """SNMP v2c community authentication

    Authenticate with SNMP v2c community

    """
    def __init__(self, community, securityname=DEFAULT_AGENT):
        SNMPAuth.__init__(self, '2c')
        self.community = community
        self.securityname = securityname
        self.auth = CommunityData(self.securityname, self.community)

    def __repr__(self):
        return '%s SNMP v2c community %s' % (self.securityname, self.community)

class SNMPv3Auth(SNMPAuth):
    """SNMP v3 community authentication

    Authenticate with SNMP v3 username, authentication password and optionally
    with separate privacy password.

    You can set authentication protocol and privacy protocol as well, supported
    protocols depend on pysnmp support for the method.

    """
    def __init__(self, username, authPass, privPass=None,
                 authProtocol=DEFAULT_V3_AUTH_PROTOCOL, privProtocol=DEFAULT_V3_PRIV_PROTOCOL):

        SNMPAuth.__init__(self, '3')
        self.username = username
        self.authPass = authPass
        self.privPass = privPass is not None and privPass or authPass

        self.auth_name = authProtocol
        self.priv_name  = privProtocol

        self.auth = UsmUserData( self.username,
            authKey=self.authPass, privKey=privPass,
            authProtocol=self.__lookup_auth_protocol(authProtocol),
            privProtocol=self.__lookup_priv_protocol(privProtocol),
        )

    def __repr__(self):
        return 'SNMP v3 auth: user %s auth %s encryption %s' % (
            self.username, self.auth_name, self.priv_name
        )

    def __lookup_auth_protocol(self,protocol):
        if not protocol:
            protocol = DEFAULT_V3_AUTH_PROTOCOL

        try:
            return V3_AUTH_PROTOCOLS[protocol.upper()]
        except KeyError:
            raise SNMPError('Unknown SNMP authentication protocol: %s' % protocol)
        except AttributeError:
            raise SNMPError('Protocol value must be a string')

    def __lookup_priv_protocol(self, protocol):
        if not protocol:
            protocol = DEFAULT_V3_PRIV_PROTOCOL

        try:
            return V3_PRIV_PROTOCOLS[protocol.upper()]
        except KeyError:
            raise SNMPError('Unknown SNMP privacy protocol: %s' % protocol)
        except AttributeError:
            raise SNMPError('Protocol value must be a string')

class SNMPIndexCache(dict):
    """Cached SNMP results

    Cache SNMP response to a file on disk

    """
    def __init__(self, path=None):
        self.path = path
        self.load()

    def load(self):
        """Load cached data

        Returns silently if file did not exist

        """
        if self.path is None or not os.path.isfile(self.path):
            return

        for (k,opts) in ConfigObj(self.path).items():
            self[k] = opts

    def save(self):
        """Save cached data

        Replace data file with current object contents

        """
        if self.path is None:
            return
        c = ConfigObj()

        for k, opts in self.items():
            if len(opts) == 0:
                continue
            if not c.has_key(k):
                c[k] = {}
            for i, data in opts.items():
                i = str(i)
                c[k][i] = data

        try:
            c.write(open(self.path, 'w'))
        except OSError, (ecode,emsg):
            raise SNMPError('Error writing index cache %s: %s' (self.path, emsg))
        except IOError, (ecode, emsg):
            raise SNMPError('Error writing index cache %s: %s' (self.path, emsg))

class SNMPClient(object):
    """
    Wrapper for SNMP SET, GET and WALK requests
    """

    def __init__(self, address, auth, port=DEFAULT_PORT, timeout=DEFAULT_TIMEOUT,
                 retries=DEFAULT_RETRIES, oid_map = {}, index_cache_path=None):

        self.logger = Logger('snmp')
        self.log = self.logger.default_stream
        self.address = address
        self.trees = {}
        self.auth_client = auth
        self.port = port
        self.timeout = timeout
        self.retries = retries

        try:
            self.target = UdpTransportTarget(
                (address, self.port),
                timeout=self.timeout,
                retries=self.retries
            )
        except socket.gaierror, emsg:
            raise SNMPError(emsg)
        except PyAsn1Error, emsg:
            raise SNMPError("ASN1 parsing error: %s" % emsg)

        self.oid_map = oid_map

        self.index_cache = SNMPIndexCache(index_cache_path)
        self.indexes = {}
        try:
            for (index, details) in self.index_cache[self.address].items():
                if details.has_key('index'):
                    details['index'] = int(details['index'])
                self.indexes[int(index)] = details

        except KeyError, emsg:
            pass

    def __repr__(self):
        return 'SNMP connection %s:%s (%s)' % (self.address, self.port, self.auth_client)

    def __getattr__(self, attr):
        if attr in self.oid_map.keys():
            config = self.oid_map[attr]
            oid = config['oid']

            try:
                return [config['decode'](k, v) for k, v in self.walk(oid).items()]
            except AttributeError:
                raise SNMPError('Invalid OID map dictionary for %s' % k)

        raise AttributeError('No such SNMPClient attribute: %s' % attr)

    def update_indexes(self):
        raise NotImplementedError('Implement update_indexes in child class')

    def set(self, oid, snmp_value):
        """
        Attempt setting the given OID to given value. Caller must prepare
        snmp_value to correct rfc1902.* reference, for example:
            rfc1902.OctetString('new name').
        """
        try:
            varBinds = tuple([
                tuple(map(lambda x:int(x), oid.lstrip('.').split('.'))),
                snmp_value
            ])
        except ValueError, emsg:
            raise SNMPError("Invalid OID: %s" % emsg)
        try:
            (e, status, index, varBinds) = CommandGenerator().setCmd(
                self.auth_client.auth,
                self.target,
                varBinds
            )

            if status != 0:
                raise SNMPError('Error setting SNMP value')

        except CarrierError, emsg:
            raise SNMPError(str(emsg))
        except socket.gaierror, emsg:
            raise SNMPError(emsg)
        except PyAsn1Error, emsg:
            raise SNMPError("ASN1 parsing error: %s" % emsg)

    def get(self, oid):
        """
        SNMP GET request to the server for given OID
        """
        try:
            oid = tuple(map(lambda x: int(x),  oid.lstrip('.').split('.')))
        except ValueError, emsg:
            raise SNMPError("Invalid OID: %s" % emsg)

        try:
            (eind, status, index, varBinds) = CommandGenerator().getCmd(
                self.auth_client.auth,
                self.target,
                oid
            )

        except CarrierError, emsg:
            raise SNMPError(str(emsg))
        except socket.gaierror, emsg:
            raise SNMPError(emsg)
        except PyAsn1Error, emsg:
            raise SNMPError("ASN1 parsing error: %s" % emsg)

        try:
            oid = varBinds[0][0]
            value = varBinds[0][1]
        except IndexError:
            raise SNMPError("No results available")

        return (oid, value)

    def walk(self, oid):
        """SNMP walk to dictionary

        Walk provided OID tree with SNMP next, return the tree as
        dictionary of (oid, value) pairs

        """
        try:
            oid = tuple(map(lambda x:int(x), oid.lstrip('.').split('.')))
        except ValueError, emsg:
            raise SNMPError("Invalid OID: %s" % emsg)
        try:
            (eind, status, index, varBinds) = CommandGenerator().nextCmd(
                self.auth_client.auth,
                self.target,
                oid
            )

        except CarrierError, emsg:
            raise SNMPError(str(emsg))
        except socket.gaierror, emsg:
            raise SNMPError(emsg)
        except PyAsn1Error, emsg:
            raise SNMPError("ASN1 parsing error: %s" % emsg)

        return dict(('.'.join(str(x) for x in v[0][0]), v[0][1]) for v in varBinds)

    def tree_key(self, oid):
        """Construct OID tree keys

        Returns a formatted OID tree key from provided string or list

        """
        try:
            if type(oid) != list:
                oid = [int(i) for i in oid.lstrip('.').split('.')]
            oid = [int(i) for i in oid]
            return '.'.join(str(i) for i in oid)

        except ValueError:
            raise SNMPError('Invalid OID: %s' % oid)

    def tree_indexes(self, oid):
        """Return OIDs mathcing prefix

        Return OID tree indexes matching provided OID from self.trees

        """
        oid = self.tree_key(oid)
        try:
            return map(lambda k: k.split('.')[-1], sorted(
                self.trees[oid].keys(),
                lambda x, y: cmp_oid(x, y)
            ))
        except KeyError:
            raise SNMPError('Tree not loaded: %s' % oid)

    def tree(self,oid, refetch=False, indexed=False):
        """Walk and store tree to self.trees

        Wrapper to self.walk to store tree results to self.trees for
        further processing

        """
        tree_id = self.tree_key(oid)
        if not refetch and self.trees.has_key(tree_id):
            tree = self.trees[tree_id]
        else:
            try:
                tree = self.walk(oid)
                self.trees[tree_id] = tree
            except CarrierError, emsg:
                raise SNMPError(str(emsg))
            except SNMPError, emsg:
                raise SNMPError(str(emsg))

        if indexed is True:
            keys = sorted(tree.keys(), lambda x, y: cmp_oid(x, y))
            return [(k.lstrip('.').split('.')[-1], tree[k]) for k in keys]
        else:
            return tree

    def fetch_trees(self, oids, refetch=False):
        """Walk and store multiple trees

        Wrapper to walk and store trees for provided OIDs to self.trees

        """
        if not isinstance(oids, list):
            raise SNMPError('Arguments oids must be a list')

        for oid in oids:
            self.tree(oid, refetch=refetch)

    def map_trees(self, oid_config):
        """Example wrappper to map trees by OIDs

        """
        oids = dict([(self.tree_key(k), oid_config[k]) for k in oid_config.keys()])

        missing = filter(lambda k: not self.trees.has_key(k), oids.keys())
        if len(missing)>0:
            raise SNMPError('Trees for OIDs not fetched: %s' % ' '.join(missing) )

        indexes = [self.tree_indexes(oid) for oid in oids.keys()]
        results = dict((int(i), {}) for i in indexes[0])
        for i in indexes[0]:
            for oid in oids.keys():
                try:
                    k = filter(lambda k: k.split('.')[-1] == i, self.trees[oid].keys())[0]
                except IndexError:
                    raise SNMPError('Error mapping OID tree %s' % oid)

                v = self.trees[oid][k]
                if results[int(i)].has_key(oid):
                    raise SNMPError('Duplicate data for %s' % oid)

                try:
                    results[int(i)][oid] = oids[oid](v)
                except ValueError:
                    results[int(i)][oid] = None

        return [(k, results[k]) for k in sorted(results.keys())]




"""
Parsers for various whois text data formats
"""

from seine.whois import WhoisError

class WhoisDataParser(dict):
	tlds = ()
	def __init__(self):
		self.domain = None
		self.data = None

	@property
	def name(self):
		return __class__

	def __repr__(self):
		return 'whois data for %s' % self.domain

	def matches_domain(self, domain):
		tld = domain.rstrip('.').split('.')[-1]
		return tld in self.tlds

	def set(self, key, value):
		if key in self.keys():
			if not isinstance(self[key], list):
				self[key] = [self[key]]
			self[key].append(value)
		else:
			self[key] = value

		return value

	def parse(self, domain, data):
		self.clear()
		self.domain = domain

		if isinstance(data, basestring):
			data = data.replace('\r\n', '\n').split('\n')
		self.data = data

		return data




import os
import sys
import re

from subprocess import check_output, CalledProcessError

from seine.address import IPv4Address, IPv6Address, \
                          EthernetMACAddress, FirewireLocalLinkAddress
from seine.classes import NetworkInterface, NetworkAddress, NetworkInterfaceList, \
                          Route, RoutingTable, ARPTable, ARPTableEntry, NetworkError


if sys.platform == 'darwin':
    NETSTAT = '/usr/sbin/netstat'
else:
    NETSTAT = '/usr/bin/netstat'


ARP_ENTRY_FORMATS = (
    re.compile('^\? \((?P<address>[0-9.]+)\) at (?P<ethernet>[^\s]+) on (?P<interface>[^\s]+) ifscope (?P<ifscope>.*)$'),
    re.compile('^\? \((?P<address>[0-9.]+)\) at (?P<ethernet>[^\s]+) on (?P<interface>[^\s]+) permanent (?P<ifscope>.*)$'),
    re.compile('^\? \((?P<address>[0-9.]+)\) at (?P<ethernet>[^\s]+) on (?P<interface>[^\s]+) expires in (?P<expires>\d+) seconds (?P<ifscope>.*)$'),
)

NETSTAT_HEADER_MATCHES = (
    re.compile('^Routing tables$'),
    re.compile('^Internet:$'),
    re.compile('^Internet6:$'),
    re.compile('^Destination\s+.*$'),
)

IPV6_ROUTE_ADDRESS_FORMATS = (
    re.compile('^(?P<address>[0-9a-f:]+)$'),
    re.compile('^(?P<address>[0-9a-f:.]+)/(?P<bitmask>\d+)$'),
    re.compile('^(?P<address>.*)%(?P<interface>[a-z0-9]+)$'),
    re.compile('^(?P<address>.*)%(?P<interface>[a-z0-9]+)/(?P<bitmask>\d+)$'),
)

def parse_ipv4_route_target(value):
    if value == 'default':
        return IPv4Address('0.0.0.0/0')

    elif value[:5] == 'link#':
        return value

    else:
        try:
            return EthernetMACAddress(value)
        except ValueError:
            pass

    return IPv4Address(value)


def parse_ipv6_route_target(value):
    if value == 'default':
        return IPv6Address('0::/0')

    elif value[:5] == 'link#':
        return value

    else:
        try:
            return EthernetMACAddress(value)
        except ValueError:
            pass

    for fmt in IPV6_ROUTE_ADDRESS_FORMATS:
        m = fmt.match(value)
        if m:
            return IPv6Address('%s/%s' % (
                m.groupdict()['address'],
                m.groupdict().get('bitmask', 128),
            ))
            break

    raise ValueError('Error parsing address from %s' % value)


class BSDNetworkAddress(NetworkAddress):
    def __init__(self, family, address, options=[]):
        NetworkAddress.__init__(self, family, address, options)

    def parse_options(self, options):
        if self.family == 'inet':
            option = None

            while options:
                if option is None:
                    option = options[0]

                elif option in ( 'netmask', 'broadcast'):
                    try:
                        value = IPv4Address(options[0])
                    except ValueError:
                        raise NetworkError('Error parsing %s %s' % (option, value))

                    if option == 'netmask':
                        self.address = IPv4Address(self.address, value)

                    if option == 'broadcast':
                        value = IPv4Address('%s/32' % value)
                        self.broadcast = value

                    option = None

                options = options[1:]

        if self.family == 'inet6':

            try:
                address, interface = self.address.split('%')
                self.address = IPv6Address(address)
                self.interface = interface

            except ValueError:
                self.address = IPv6Address(self.address)
                self.interface = None

            option = None
            while options:
                option = options[0]

                if option == 'prefixlen':
                    self.address = IPv6Address('%s/%s' % (self.address.address, options[1]))
                    options = options[1:]

                elif option == 'scopeid':
                    self.flags['scope_id'] = options[1]
                    options = options[1:]

                elif option in ( 'deprecated', 'autoconf', 'temporary', ):
                    if option not in self.flags.keys():
                        self.flags[option] = True

                options = options[1:]


class BSDNetworkInterface(NetworkInterface):
    def __init__(self, name):
        NetworkInterface.__init__(self, name)


    def __parse_status__(self, fields):
        value = ' '.join(fields)

        if value == 'active':
            self.media_status = True

        elif value == 'inactive':
            self.media_status = False

        else:
            raise ValueError('Unsupported status field value: %s' % value)

    def __parse_options__(self, fields):
        index, options = fields.rstrip('>').split('<',1)
        for key in options.split(','):
            self.options[key] = True

    def __parse_nd6_options__(self, fields):
        index, options = fields.rstrip('>').split('<',1)
        for key in options.split(','):
            self.nd6_options[key] = True

    def __parse_media__(self, fields):
        if fields[0] == 'autoselect' or fields[:2] == ['Ethernet', 'autoselect']:
            self.media_config = ' '.join(fields[1:2])
            if len(fields) == 1:
                media_type = None
            elif fields[1] == '(none)':
                self.media_state = None
            else:
                states = [x.lstrip('(').rstrip(')').lstrip('<').rstrip('>') for x in fields[1:]]
                self.media_state = states

        elif ' '.join(fields) == '<unknown type>':
            self.media_config = None
            self.media_state = None

        else:
            raise ValueError('Error parsing media type %s' % ' '.join(fields))

    def parse_flags(self, flags):
        fields = flags.split()
        while fields:
            if fields[0] == 'mtu':
                self.mtu = fields[1]
                fields = fields[1:]

            if fields[0][:6] == 'flags=':
                number, flags = fields[0][6:].split('<', 1)
                self.flags.extend(flags.replace('>','').split(','))

            fields = fields[1:]

    def parse_bridge_configuration(self, fields):
        if fields[0] == 'Configuration:':
            return

        if fields[0] == 'id':
            self.configuration['bridge_id'] = fields[1]
            self.configuration['bridge_priority'] = int(fields[3])
            self.configuration['bridge_hello_time'] = int(fields[5])
            self.configuration['bridge_fwd_delay'] = int(fields[7])

        elif fields[0] == 'member:':
            if 'bridge_member_interfaces' not in self.configuration:
                self.configuration['bridge_member_interfaces'] = {}

            config = {}
            if fields[2][:6] == 'flags=':
                index, flags = fields[2][6:].split('<',1)
                config['index'] = index
                config['flags'] = flags.replace('>', '').split(',')
                self.configuration['bridge_member_interfaces'][fields[1]] = config

        elif fields[0] == 'maxage':
            self.configuration['bridge_max_age'] = int(fields[1])
            self.configuration['bridge_hold_count'] = int(fields[3])
            self.configuration['bridge_proto'] = fields[5]
            self.configuration['bridge_max_addr'] = int(fields[7])
            self.configuration['bridge_timeout'] = int(fields[9])

        elif fields[0] == 'ifmaxaddr':
            self.configuration['bridge_ifmaxaddr'] = int(fields[1])
            self.configuration['bridge_ifmaxaddr_port'] = int(fields[3])
            self.configuration['bridge_ifmaxaddr_priority'] = int(fields[5])
            self.configuration['bridge_ifmaxaddr_path_cost'] = int(fields[8])

        elif fields[0] == 'ipfilter':
            self.configuration['bridge_ipfilter_status'] = fields[1]
            self.configuration['bridge_ipfilter_flags'] = fields[3]

        elif ' '.join(fields[:2]) == 'root id':
            self.configuration['bridge_root_id'] = fields[2]
            self.configuration['bridge_root_priority'] = int(fields[4])
            self.configuration['bridge_root_ifcost'] = int(fields[6])
            self.configuration['bridge_root_port'] = int(fields[8])

    def parse(self, line):
        line = line.lstrip()
        fields = line.split()

        if fields[0] == 'inet':
            self.ipv4_addresses.append(BSDNetworkAddress('inet', fields[1], fields[2:]))

        elif fields[0] == 'inet6':
            self.ipv6_addresses.append(BSDNetworkAddress('inet6', fields[1], fields[2:]))

        elif fields[0] == 'ether':
            self.media = EthernetMACAddress(fields[1])

        elif fields[0] == 'lladdr':
            self.media = FirewireLocalLinkAddress(fields[1])

        elif fields[0] == 'status:':
            self.__parse_status__(fields[1:])

        elif fields[0] == 'media:':
            self.__parse_media__(fields[1:])

        elif fields[0][:8] == 'options=':
            self.__parse_options__(fields[0][8:])

        elif fields[0] == 'nd6':
            self.__parse_nd6_options__(fields[1])

        elif self.name[:6] == 'bridge':
            self.parse_bridge_configuration(fields)

        elif fields[0] == 'vlan:':
            self.vlan_id = fields[1]
            self.vlan_device = fields[4]


class BSDNetworkInterfaces(NetworkInterfaceList):
    def update(self):
        self.__delslice__(0, len(self))

        try:
            output = check_output(['/sbin/ifconfig'])
        except CalledProcessError, emsg:
            raise NetworkError('Error getting interfaces: %s' % emsg)

        interface = None
        for l in [l for l in output.split('\n') if l.strip() != '']:
            if l.startswith('\t'):
                interface.parse(l)

            else:
                name, flags = l.split(':', 1)
                interface = BSDNetworkInterface(name)
                interface.parse_flags(flags)
                self.append(interface)


class IPv4RouteEntry(Route):
    def __init__(self, table, line):
        fields = line.split()

        address = parse_ipv4_route_target(fields[0])
        target = parse_ipv4_route_target(fields[1])
        flags = fields[2]
        refs = int(fields[3])
        use = int(fields[3])
        interface = fields[4]

        if len(fields) == 6:
            expires = fields[5]
        else:
            expires = None

        Route.__init__(self, table, address, target, flags=flags, refs=refs, use=use, interface=interface, expires=expires)

    def __str__(self):
        return '%s %s' % (self.address, self.target)


class IPv6RouteEntry(Route):
    def __init__(self, table, line):
        fields = line.split()

        address = parse_ipv6_route_target(fields[0])
        target = parse_ipv6_route_target(fields[1])
        flags = fields[2]
        interface = fields[3]

        if len(fields) == 5:
            expires = fields[4]
        else:
            expires = None

        Route.__init__(self, table, address, target, flags=flags, interface=interface, expires=expires)

    def __str__(self):
        return '%s %s' % (self.address, self.target)


class BSDRoutes(RoutingTable):
    """BSD style netstat parsing

    Parse BSD style netstat -rn outputs

    """
    def update(self):
        self.ipv4 = []
        self.ipv6 = []

        try:
            output = check_output([NETSTAT, '-rn', '-f', 'inet'])
        except CalledProcessError, emsg:
            raise NetworkError('Error getting IPv4 routes: %s' % emsg)

        interface = None
        for l in [l for l in output.split('\n') if l.strip() != '']:
            if l.strip() == '':
                continue
            header = False
            for m in NETSTAT_HEADER_MATCHES:
                if m.match(l):
                    header = True
                    break
            if header:
                continue

            self.ipv4.append(IPv4RouteEntry(self, l))

        try:
            output = check_output([NETSTAT, '-rn', '-f', 'inet6'])
        except CalledProcessError, emsg:
            raise NetworkError('Error getting IPv6 routes: %s' % emsg)

        interface = None
        for l in [l for l in output.split('\n') if l.strip() != '']:
            if l.strip() == '':
                continue
            header = False
            for m in NETSTAT_HEADER_MATCHES:
                if m.match(l):
                    header = True
                    break
            if header:
                continue

            self.ipv6.append(IPv6RouteEntry(self, l))

        self.sort()

    def sort(self):
        self.ipv4.sort()
        self.ipv6.sort()


class BSDARP(ARPTable):
    def update(self):
        self.__delslice__(0, len(self))

        try:
            output = check_output(['/usr/sbin/arp', '-an'])
        except CalledProcessError, emsg:
            raise NetworkError('Error getting ARP table: %s' % emsg)

        for l in [l for l in output.split('\n') if l.strip() != '']:
            if l.strip() == '':
                continue

            for fmt in ARP_ENTRY_FORMATS:
                m = fmt.match(l)
                if m:
                    self.append(ARPTableEntry(**m.groupdict()))




#!/usr/bin/env python
"""
Module for abstract resolver queries with dns-python to easy to use
functions.

This is used mainly to query arbitrary servers for all values, not
just one of the values like with DNS lookups, and to return the data
as an easily understandable dictionary.
"""

import socket

from seine.dns import DNSError
import dns.resolver, dns.query, dns.rdatatype
import dns.rdtypes.ANY, dns.rdtypes.IN

class QueryError(Exception):
    pass

# Map dns.rdtypes subclass pahts to attributes to fetch from the response
# class name
RDTYPE_NAME_MAP = {
    'dns.rdtypes.ANY.SOA':   [
        'serial', 'retry', 'refresh', 'expire', 'minimum', 'mname', 'rname',
    ],
    'dns.rdtypes.ANY.NS':    ['target'],
    'dns.rdtypes.IN.A':      ['address'],
    'dns.rdtypes.IN.AAAA':   ['address'],
    'dns.rdtypes.ANY.PTR':   ['target'],
    'dns.rdtypes.ANY.CNAME': ['target'],
    'dns.rdtypes.ANY.MX':    ['preference', 'exchange'],
    'dns.rdtypes.ANY.TXT':   ['strings'],
    'dns.rdtypes.ANY.SSHFP': ['algorithm', 'fingerprint'],
}
# Initialize a dictionary with the class references from dns.rdtypes
RDTYPE_DICT_MAP = {}
for path in RDTYPE_NAME_MAP.keys():
    name = path.split('.')[-1]
    m = __import__(path, globals(), fromlist=[name])
    nk = getattr(m, name)
    RDTYPE_DICT_MAP[nk] = RDTYPE_NAME_MAP[path]

def rrname_mapped(value):
    """
    Checks if given rrtype value is supported (mapping from dns module
    data structures is implemented). Returns the rrname in uppercase or
    raises ValueError if rrtype is not implemented.
    """
    rrname = value.upper()
    if rrname not in map(lambda x: x.split('.')[-1], RDTYPE_NAME_MAP.keys()):
        raise ValueError('RR name %s not yet supported.' % rrname)

    return rrname

def resolve_records(query, nameserver, rrtype='A', timeout=5):
    """
    Try resolving given query (for. example www.google.com) from given
    nameservers, querying given rrtype.

    Returns each response with given RR type as a dictionary, containing
    details relevant for the query in question. Returns empty list if no
    details could be found.
    """

    try:
        rrname = rrname_mapped(rrtype)
    except ValueError, e:
        raise ValueError(e)

    try:
        rdtype = getattr(dns.rdatatype, rrname)
    except AttributeError:
        raise ValueError('Unsupported RR name %s' % rrname)

    try:
        timeout = int(timeout)
    except ValueError:
        raise ValueError('Invalid query timeout value: %s' % timeout)

    try:
        response = dns.query.udp(
            q=dns.message.make_query(query, rdtype),
            where=nameserver, timeout=timeout
        )
    except socket.error, (ecode, emsg):
        raise QueryError('Nameserver %s: %s' % (nameserver, emsg))
    except dns.exception.Timeout:
        raise QueryError('Timeout resolving from %s' % nameserver)

    authority = []
    for rs in response.authority:
        for r in rs:
            try:
                attrs = RDTYPE_DICT_MAP[type(r)]
            except KeyError:
                raise ValueError('No mapping for %s' % r.__class__)

            entry = dict([(k, getattr(r, k)) for k in attrs])
            entry['rdclass'] = dns.rdataclass.to_text(r.rdclass)
            entry['rrtype'] = dns.rdatatype.to_text(r.rdtype)

            authority.append(entry)

    additional = []
    for rs in response.additional:
        for r in rs:
            try:
                attrs = RDTYPE_DICT_MAP[type(r)]
            except KeyError:
                raise ValueError('No mapping for %s' % r.__class__)

            entry = dict([(k, getattr(r, k)) for k in attrs])
            entry['name'] = rs.name
            entry['rdclass'] = dns.rdataclass.to_text(r.rdclass)
            entry['rrtype'] = dns.rdatatype.to_text(r.rdtype)
            additional.append(entry)

    results = []
    for rs in filter(lambda rs: rs.rdtype == rdtype, response.answer):
        for r in rs:
            try:
                attrs = RDTYPE_DICT_MAP[type(r)]
            except KeyError:
                raise ValueError('No mapping for %s' % r.__class__)

            entry = dict([(k, getattr(r, k)) for k in attrs])
            entry['name'] = rs.name
            entry['rdclass'] = dns.rdataclass.to_text(r.rdclass)
            entry['rrtype'] = dns.rdatatype.to_text(r.rdtype)
            results.append(entry)

    return {
        'results': results,
        'authority': authority,
        'additional': additional,
    }




"""
Wrapper for urllib2 to do some basic tasks with URLs.
Written by Ilkka Tuohela <hile@iki.fi>, 2007-2012.
Licensed under BSD license.

NOTE:
This module is a obsolete hack: please use 'requests' module.

"""

import os
import sys
import re
import socket
import urllib
import urllib2
import cookielib

from seine.address import IPv4Address, IPv6Address

# Default socket timeout for requests,  can be overriden with 'timeout' keyword
DEFAULT_RETRIES = 1
DEFAULT_TIMEOUT = 30
DEFAULT_USERAGENT='Mozilla/5.0 (Windows; U; Windows NT 6.0; en-GB; rv:1.9.0.8) Gecko/2009032609 Firefox/3.0.8'
DEFAULT_PROTOCOL = 'http'

class HTTPRequestError(Exception):
    def __str__(self):
        return ' '.join(str(x) for x in self.args)

class HTTPRequest(object):
    """
    URL requests
    Arguments:
    proxy_url   URL to proxy, or environment http_proxy
    user_agent  User-Agent header or DEFAULT_USERAGENT
    """
    def __init__(self, **kwargs):
        self.req = None
        self.headers = {
            'User-Agent': kwargs.get('user_agent', DEFAULT_USERAGENT),
        }
        self.cookies = cookielib.LWPCookieJar()

        self.proxy_url= kwargs.get('http_proxy', os.getenv('http_proxy'))
        self.timeout = kwargs.get('timeout', DEFAULT_TIMEOUT)
        self.retries = kwargs.get('retries', DEFAULT_RETRIES)
        self.auth_realm = kwargs.get('auth_realm', None)
        self.auth_user = kwargs.get('auth_user', None)
        self.auth_pass = kwargs.get('auth_pass', None)

        if self.auth_realm is not None:
            self.auth = urllib2.ProxyBasicAuthHandler()
            self.auth.add_password(
                self.auth_realm,
                self.proxy_url,
                self.auth_user,
                self.auth_pass
            )
        else:
            self.auth = None

    def __getattr__(self, attr):
        if attr == 'url':
            return self.get_url()
        if attr == 'proxy_handler':
            return self.get_proxy_handler()
        if attr == 'opener':
            return self.get_opener()
        raise AttributeError('No such HTTPRequest attribute: %s' % attr)

    @property
    def get_url(self):
        return self.req is not None and self.req.get_full_url() or None

    def get_proxy_handler(self):
        if self.proxy_url is not None:
            handler = (
                urllib2.ProxyHandler(
                    {'http':self.proxy_url, 'https':self.proxy_url
                }),
                self.auth
            )
        else:
            handler = (urllib2.ProxyHandler(), None)
        return handler

    def get_opener(self):
        (proxy, auth) = self.get_proxy_handler()
        if auth:
            return urllib2.build_opener(auth, proxy)
        else:
            return urllib2.build_opener(proxy)

    def request(self, url, method='GET', **kwargs):
        """
        Send a HTTP GET or POST request.
        Returns tuple (code, data, headers)
        """
        old_timeout = socket.getdefaulttimeout()
        socket.setdefaulttimeout(kwargs.get('timeout', self.timeout))

        h = dict(self.headers)
        h.update(kwargs.get('headers', {}))
        form = kwargs.get('form', None)
        data = form is not None and urllib.urlencode(form) or None

        try:
            retries = 0
            opener = self.get_opener()
            while retries<=self.retries:
                retries+=1
                try:
                    req = opener.open(urllib2.Request(url, data, h))
                except urllib2.URLError, e:
                    raise HTTPRequestError(str(e))
                try:
                    code = 200
                    data = req.read()
                    headers = req.info()
                except urllib2.HTTPError, e:
                    code = e.code
                    data = e.read()
                    headers = {}
                except urllib2.URLError, e:
                    msg = e.reason[1]
                    if self.proxy_url:
                        msg += ' (using proxy %s)' % self.proxy_url
                    raise HTTPRequestError(msg)
                except socket.timeout:
                    if retries >= self.retries:
                        raise HTTPRequestError('Request timeout')
                return (code, data, headers)
        finally:
            socket.setdefaulttimeout(old_timeout)

    def GET(self, url, headers={}, timeout=None):
        return self.request(url, method='GET', form=None,
            headers=headers, timeout=timeout
        )

    def POST(self, url, form):
        return self.request(url, method='POST', form=form,
            headers=headers, timeout=timeout
        )




"""
Proxy autoconfiguration processing, for both passing autoconfig URL
and retriving from WPAD DNS configured hosts.
"""

import requests
import urllib2
import socket
import StringIO

try:
    import pacparser
except ImportError:
    raise ImportError('Module pacparser is not installed.')

DEFAULT_QUERY_TIMEOUT = 3.0

class ProxyConfigError(Exception):
    pass


class ProxyAutoConfig(object):
    """Wrapper class for pacparser

    Wrap pacparser URL handling to a class

    """

    def __init__(self, pac_url=None, timeout=DEFAULT_QUERY_TIMEOUT):
        if pac_url is None:
            default_timeout = socket.getdefaulttimeout()
            socket.setdefaulttimeout(timeout)
            try:
                wpad_domain = '.'.join(socket.getfqdn('wpad').split('.')[1:])
            except socket.GAIError, emsg:
                raise ProxyConfigError('Error querying wpad hostname: %s' % emsg)
            socket.setdefaulttimeout(default_timeout)

            if wpad_domain == '':
                raise ProxyConfigError('No URL given and WPAD hostname not available')
            else:
                pac_url = 'http://wpad.%s/wpad.dat' % wpad_domain

        self.pac_url = pac_url
        self.data = None

    def __read_pac__(self):
        """Read PAC URL

        Read and parse wpad.dat data from configured URL

        """
        if self.pac_url is None:
            raise ProxyConfigError('PAC URL not configured')

        res = requests.get(self.pac_url)
        if res.status_code != 200:
            raise ProxyConfigError('Error retrieving PAC from %s' % self.pac_url)

        self.data = StringIO.StringIO()
        self.data.write(res.content)
        self.data.seek(0)

    def __repr__(self):
        return 'PAC: %s' % self.pac_url

    def __str__(self):
        """PAC as string

        Return PAC contents as string

        """

        if self.data is None:
            self.__read_pac__()

        self.data.seek(0)
        return self.data.read()

    def validate(self):
        """Validate PAC with pacparser

        Validate resolved PAC configuration with pacparser libarry

        """
        pacparser.init()
        pacparser.parse_pac_string(str(self))
        pacparser.cleanup()

    def find_proxies(self, url):
        """Parse proxy URL for provided URL

        Parse PAC and return proxy URL for given URL

        """

        try:
            protocol = url.split('://')[0]
        except ValueError:
            raise ProxyConfigError('Invalid URL: %s' % url)

        try:
            pacparser.init()
            pacparser.parse_pac_string(str(self))
            proxies = pacparser.find_proxy(url)
            pacparser.cleanup()
        except:
            raise ProxyConfigError('Error parsing PAC: %s' % self.pac_url)
        data = {}
        for v in [x.strip() for x in proxies.split(';')]:
            if v == 'DIRECT':
                continue

            if v[:6] == 'PROXY ':
                data[protocol] = v[6:]
        return data

    def proxy_handler(self, url):
        """Create proxyhandler for URL

        Create a urllib2 proxy handler from parser data for provided url

        """

        try:
            protocol = url.split('://')[0]
        except ValueError:
            raise ProxyConfigError('Invalid URL: %s' % url)

        proxies = self.find_proxies(url)

        if proxies:
            return urllib2.ProxyHandler(proxies)
        else:
            return urllib2.ProxyHandler()




#!/usr/bin/env python
"""
Nagios test module to query whois servers for a domain's data,
use by libexec/check_whois script to check expiration status.
"""

import sys
import os
import logging
import time
import socket
import select

from seine.address import IPv4Address, IPv6Address
from seine.whois import WhoisError
from seine.whois.cache import WhoisServerCache
from seine.whois.parsers import WhoisData

WHOIS_PORT = 43
WHOIS_SERVER_TIMEOUT = 15
WHOIS_BUFFER_SIZE = 1024

TLD_REQUIRES_EQUALS = ['com']

logger = logging.getLogger()

class WhoisClient(object):
    def __init__(self):
        self.cache = WhoisServerCache()

    def query(self, domain):
        data = WhoisData(domain)
        data.query(self.cache.query(domain))
        try:
            self.cache.save()
        except WhoisError, emsg:
            logger.debug(emsg)
            pass

        return data
